package djmil.cordacheckers.gameresult;

import java.security.PublicKey;
import java.time.Duration;
import java.time.Instant;
import java.util.UUID;

import djmil.cordacheckers.VNode;
import djmil.cordacheckers.contracts.GameCommand;
import djmil.cordacheckers.gamestate.CommitTrx;
import djmil.cordacheckers.gamestate.GetFlow;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameResultState;
import djmil.cordacheckers.states.GameState;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.flows.SubFlow;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.membership.MemberInfo;

public class GameResultCommiter implements SubFlow<SecureHash> {

    private final UUID gameUuid;
    private final GameCommand command;

    public GameResultCommiter(UUID gameUuid,  GameCommand command) {
        this.gameUuid = gameUuid;
        this.command = command;
    }

    @CordaInject
    public FlowEngine flowEngine;
    
    @CordaInject
    public MemberLookup memberLookup;

    @CordaInject
    public UtxoLedgerService utxoLedgerService;

    @Override
    @Suspendable
    public SecureHash call() {
        final MemberInfo custodianInfo = findCustodian();

        final StateAndRef<GameState> gameBoardSar = this.flowEngine
                .subFlow(new GetFlow(this.gameUuid));
        final GameBoardState gameBoard = (GameBoardState)gameBoardSar.getState().getContractState();

        final GameResultState gameResult = gameResultBuilder(gameBoard, custodianInfo);

        final UtxoSignedTransaction gameResultTrx = utxoLedgerService.createTransactionBuilder()
            .addCommand(this.command)
            .addInputState(gameBoardSar.getRef())
            .addOutputState(gameResult)
            .addSignatories(gameResult.getParticipants())
            .setNotary(gameBoardSar.getState().getNotaryName())
            .setTimeWindowUntil(Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
            .toSignedTransaction();
System.out.println("GameResultCommiter");
        return this.flowEngine.subFlow( 
            new CommitTrx(gameResultTrx, 
                    getCounterparty(gameResult),
                    custodianInfo.getName()) );
    }

    @Suspendable
    MemberInfo findCustodian() {
        return memberLookup.lookup()
            .stream()
            .filter(member -> VNode.isCordaCherckersCustodian(member) )
            .reduce((a,b) -> {throw new IllegalStateException("Multiple Custodian VNodes");})
            .orElseThrow( () -> new IllegalStateException("No Custodian VNode found"));
    }

    @Suspendable 
    GameResultState gameResultBuilder(GameBoardState gameBoard, MemberInfo custodiaInfo) {
        final PublicKey custodianPublicKey = custodiaInfo.getLedgerKeys().get(0);
        final MemberX500Name myName = memberLookup.myInfo().getName();
        
        switch(this.command.getAction()) {
            case CLAIM_VICTORY:
                return new GameResultState(myName, // i'm a winner
                            gameBoard, custodianPublicKey);

            case SURRENDER:
                return new GameResultState(gameBoard.getOpponentName(myName), // me surrender to opponent
                            gameBoard, custodianPublicKey);

            case DRAW_ACCEPT:
                return new GameResultState(null, // there is no winner, it's a draw
                            gameBoard, custodianPublicKey);

            default:
                throw new IllegalStateException("GameResult: bad reason");
        } 
    }

    @Suspendable
    MemberX500Name getCounterparty(GameState gameState) {
        final MemberX500Name myName = this.memberLookup.myInfo().getName();
        return gameState.getOpponentName(myName);
    }

}
